#!/usr/bin/env python3.6
# -*- Coding: UTF-8 -*-
"""
Roleta de Seleção de Indivíduos para cruzamento.

Programa sob licença GNU V.3.
Desenvolvido por: E. S. Pereira.
Versão 0.0.1.
"""

from numpy.random import random
from numpy import array

from .selecao import Selecao


class Roleta(Selecao):
    """
    Seleciona indivíduos para cruzamento usando
    roleta de seleção.
    Recebe como entrada:
        populacao - Objeto criado a partir da classe Populacao.
    """
    def __init__(self, populacao):
        super(Roleta, self).__init__(populacao)

    def selecionar(self, fitness):
        """Roleta de seleção de indivíduos."""
        if fitness is None:
            fitness = self.populacao.avaliar()
        fmin = fitness.min()
        fitness = fitness - fmin
        total = fitness.sum()
        parada = total * (1.0 - random())
        parcial = 0
        i = 0
        while True:
            if i > fitness.size - 1:
                break
            parcial += fitness[i]
            if parcial >= parada:
                break
            i += 1
        return i - 1



#!/usr/bin/env python3.6
# -*- Coding: UTF-8 -*-
"""
Cruzamento via embaralhamento.

Programa sob licença GNU V.3.
Desenvolvido por: E. S. Pereira.
Versão 0.0.1.
"""

from numpy.random import shuffle, randint
from .cruzamento import Cruzamento, NoCompatibleIndividualSize


class Embaralhamento(Cruzamento):
    """
    Gerador de população via embaralhanmento e cruzamento de um ponto.

    Entrada:
        tamanho_populacao - Tamanho final da população resultante.
    """
    def __init__(self, tamanho_populacao):
        super(Embaralhamento, self).__init__(tamanho_populacao)

    def cruzamento(self, progenitor1, progenitor2):
        """
        Cruzamento de dois indivíduos via embaralhanmento um pontos.

        Entrada:
            ind1 - Primeiro indivíduo
            ind2 - Segundo indivíduo
        O tamanho de ambos os indivíduos deve ser igual, do contrário um erro
        será levantado.
        """
        n1 = len(progenitor1)
        n2 = len(progenitor2)

        if n1 != n2:
            msg = "Tamanho ind1 {0} diferente de ind2 {1}".format(n1, n2)
            raise NoCompatibleIndividualSize(msg)

        order = list(range(n1))
        shuffle(order)

        ponto = randint(1, n1 - 1)
        desc1 = progenitor1.copy()
        desc2 = progenitor2.copy()

        desc1[:] = desc1[order]
        desc2[:] = desc2[order]

        desc1[ponto:], desc2[ponto:] =  desc2[ponto:], desc1[ponto:]

        tmp1 = desc1.copy()
        tmp2 = desc2.copy()

        for i, j in enumerate(order):
            desc1[j] = tmp1[i]
            desc2[j] = tmp2[i]

        return desc1, desc2



#!/usr/bin/env python3.6
# -*- Coding: UTF-8 -*-
"""
Evolução.

Programa sob licença GNU V.3.
Desenvolvido por: E. S. Pereira.
Versão 0.0.1.
"""

from numpy.random import random


class Evolucao:
    """
    Usando operadores genéticos, coloca uma população para evoluir.

    Entrada:
        populacao - Objeto do tipo Populacao
        selecao - Objeto do tipo Selecao
        cruzamento - Objeto do tipo cruzamento
        mutacao - Objeto do tipo  Mutacao.
    """

    def __init__(self, populacao, selecao, cruzamento, mutacao):
        self.populacao = populacao
        self.selecao = selecao
        self.cruzamento = cruzamento
        self.mutacao = mutacao
        self._melhor_solucao = None
        self._geracao = 0
        self._nsele = None
        self._pcruz = None
        self._epidemia = None
        self._manter_melhor = True
        self._fitness = None
        self._first = True

    def _set_epidemia(self, epidemia):
        self._epidemia = epidemia

    def _set_manter_melhor(self, manter):
        self._manter_melhor = manter

    def _get_manter_melhor(self):
        return self._manter_melhor

    def _get_epidemia(self):
        return self._epidemia

    def _set_nsele(self, nsele):
        self._nsele = nsele

    def _get_nsele(self):
        return self._nsele

    def _set_pcruz(self, pcruz):
        self._pcruz = pcruz

    def _get_pcruz(self):
        return self._pcruz

    def _get_first(self):
        return self._first

    def _set_first(self, first):
        self._first = first

    @property
    def melhor_solucao(self):
        return self._melhor_solucao

    @property
    def geracao(self):
        return self._geracao

    def evoluir(self):
        """
        Evolução elitista, por uma geração, da popução.
        """
        if self._first is True:
            self._fitness = self.populacao.avaliar()
            self._first = False

        self._melhor_solucao = self.populacao.populacao[-1].copy()

        subpopulacao = self.selecao.selecao(self._nsele, fitness=self._fitness)
        populacao = self.cruzamento.descendentes(subpopulacao, pcruz=self._pcruz)

        self.mutacao.populacao = populacao
        self.mutacao.mutacao()
        self.populacao.populacao[:] = populacao[:]

        self._geracao += 1

        if self._epidemia is not None:
            if self._geracao % self._epidemia == 0 and random() < 0.8:
                """Passo Epidêmico"""
                print("Epidemia")
                self._possivel_local = 0
                self.populacao.gerar_populacao()

        if self._manter_melhor is True:
            self.populacao.populacao[0] = self._melhor_solucao

        self._fitness = self.populacao.avaliar()

        #if self._manter_melhor is True:
        #    self._melhor_solucao = self.populacao.populacao[-1].copy()


        return self._fitness.min(), self._fitness.max()

    nsele = property(_get_nsele, _set_nsele)
    pcruz = property(_get_pcruz, _set_pcruz)
    epidemia = property(_get_epidemia, _set_epidemia)
    manter_melhor = property(_get_manter_melhor, _set_manter_melhor)
    first = property(_get_first, _set_first)



#!/usr/bin/env python3.6
# -*- Coding: UTF-8 -*-
"""
Mutação flip.

Programa sob licença GNU V.3.
Desenvolvido por: E. S. Pereira.
Versão 0.0.1.
"""

from numpy.random import randint
from numpy import array, newaxis

from .mutacao import Mutacao


class SequenciaReversa(Mutacao):
    """
    Mutaçao flip.

    Entrada:
        populacao - vetor de população que deverá sofrer mutação.
        pmut - probabilidade de ocorrer uma mutação.
    """
    def __init__(self, pmut):
        super(SequenciaReversa, self).__init__(pmut)

    def mutacao(self):
        """
        Alteração genética de membros da população usando sequência reversa.
        """
        nmut = self.selecao()
        if nmut.size != 0:
            for k in nmut:
                i = randint(0, self.ngen - 1)
                j = randint(0, self.ngen - 1)
                while i == j:
                    j = randint(0, self.ngen - 1)
                if i > j:
                    i, j = j, i
                self.populacao[k, i:j] = self.populacao[k, i:j][::-1]



#!/usr/bin/env python3.6
# -*- Coding: UTF-8 -*-
"""
Cruzamento.

Programa sob licença GNU V.3.
Desenvolvido por: E. S. Pereira.
Versão 0.0.1.
"""

from numpy.random import randint, random
from numpy import array


class NoCompatibleIndividualSize(Exception):
    pass



class Cruzamento:
    """
    Classe abstrata representando o cruzamento.

    Entrada:
        tamanho_populacao - Tamanho final da população resultante.
    """
    def __init__(self, tamanho_populacao):
        self.tamanho_populacao = tamanho_populacao

    def cruzamento(self, progenitor1, progenitor2):
        raise NotImplementedError("A ser implementado")


    def descendentes(self, subpopulacao, pcruz):
        """
        Retorna uma nova população de tamanho tamanho_populacao
        através do cruzamento usando k-pontos.

        Entrada:
            subpopulacao - Vetor contendo indivíduos para serem selecionados
                           para cruzamento.
            pcruz - probabilidade de cruzamento entre dois indivíduos
                    selecionados.
        """
        nova_populacao = []
        npop = len(subpopulacao)

        while len(nova_populacao) < self.tamanho_populacao - 1:
            i = randint(0, npop - 1)
            j = randint(0, npop - 1)
            while j == i:
                j = randint(0, npop - 1)

            cruzar = random()
            if cruzar < pcruz:
                desc1, desc2 = self.cruzamento(subpopulacao[i], subpopulacao[j])
                nova_populacao.append(desc1)
                nova_populacao.append(desc2)

        return array(nova_populacao)



#!/usr/bin/env python3.6
# -*- Coding: UTF-8 -*-
"""
Mutação flip.

Programa sob licença GNU V.3.
Desenvolvido por: E. S. Pereira.
Versão 0.0.1.
"""

from numpy.random import randint
from numpy import array

from .mutacao import Mutacao


class Flip(Mutacao):
    """
    Mutaçao flip.

    Entrada:
        populacao - vetor de população que deverá sofrer mutação.
        pmut - probabilidade de ocorrer uma mutação.
    """
    def __init__(self, pmut):
        super(Flip, self).__init__(pmut)

    def mutacao(self):
        """Alteração genética de membros da população usando mutação flip."""
        nmut = self.selecao()
        if nmut.size != 0:
            genflip = array([randint(0, self.ngen - 1) for _ in nmut])
            self.populacao[nmut, genflip] = 1-self.populacao[nmut, genflip]



#!/usr/bin/env python3.6
# -*- Coding: UTF-8 -*-
"""
Mutação.

Programa sob licença GNU V.3.
Desenvolvido por: E. S. Pereira.
Versão 0.0.1.
"""

from numpy.random import randint, random
from numpy import array


class Mutacao:
    """
    Classe base para operadores de mutação:
    Entrada:
        pmut - probabilidade de ocorrer uma mutação.
    """

    def __init__(self, pmut):
        self.pmut = pmut
        self._populacao = None
        self.npop = None
        self.ngen = None

    def _set_populacao(self, populacao):
        self._populacao = populacao
        self.npop = self._populacao.shape[0]
        self.ngen = self._populacao.shape[1]

    def _get_populacao(self):
        return self._populacao

    def selecao(self):
        nmut = array([i for i in range(self.npop) if random() < self.pmut])
        return nmut

    def mutacao(self):
        raise NotImplementedError("A ser implementado")

    populacao = property(_get_populacao, _set_populacao)



#!/usr/bin/env python3.6
# -*- Coding: UTF-8 -*-
"""
Gerador aleatório de população.

Programa sob licença GNU V.3.
Desenvolvido por: E. S. Pereira.
Versão 0.0.1.
"""

from numpy.random import randint
from numpy import argsort, unique


class Populacao:
    """
    Cria e avalia uma população.
    Recebe como entrada:
        avaliacao - Função que recebe um indivíduo como entrada e retorna
                    um valor numérico.
        cromossos_totais - Numero inteiro representando o tamanho da cadeia
                           genética do indivíduo.
        tamanho_populacao - Numero inteiro representando o número total de
                            indivíduos na população.
    """

    def __init__(self, avaliacao, genes_totais, tamanho_populacao):
        self.avaliacao = avaliacao
        self.genes_totais = genes_totais
        self.tamanho_populacao = tamanho_populacao
        self.gerar_populacao()

    def gerar_populacao(self):
        """Gerador aleatório de população."""
        self.populacao = randint(0, 2, size=(self.tamanho_populacao,
                                             self.genes_totais),
                                 dtype='b')

    def avaliar(self):
        """Avalia e ordena a população."""
        u, indices = unique(self.populacao, return_inverse=True, axis=0)
        valores = self.avaliacao(u)
        valores = valores[indices]
        ind = argsort(valores)

        self.populacao[:] = self.populacao[ind]
        valores = valores[ind]
        return valores[ind]



#!/usr/bin/env python3.6
# -*- Coding: UTF-8 -*-
"""
Mutação dupla troca.

Programa sob licença GNU V.3.
Desenvolvido por: E. S. Pereira.
Versão 0.0.1.
"""

from numpy.random import randint
from numpy import array

from .mutacao import Mutacao


class DuplaTroca(Mutacao):
    """
    Mutaçao dupla troca.

    Entrada:
        populacao - vetor de população que deverá sofrer mutação.
        pmut - probabilidade de ocorrer uma mutação.
    """
    def __init__(self, pmut):
        super(DuplaTroca, self).__init__(pmut)

    def mutacao(self):
        """Alteração genética de membros da população usando dupla troca."""
        nmut = self.selecao()
        if nmut.size != 0:
            gen1 = array([randint(0, self.ngen - 1) for _ in nmut])
            gen2 = array([randint(0, self.ngen - 1) for _ in nmut])

            self.populacao[nmut, gen1], self.populacao[nmut, gen2] = \
            self.populacao[nmut, gen2], self.populacao[nmut, gen1]



#!/usr/bin/env python3.6
# -*- Coding: UTF-8 -*-
"""
Mutação dupla troca.

Programa sob licença GNU V.3.
Desenvolvido por: E. S. Pereira.
Versão 0.0.1.
"""

from numpy.random import randint
from numpy import array

from .mutacao import Mutacao


class NTrocas(Mutacao):
    """
    Mutaçao dupla troca.

    Entrada:
        populacao - vetor de população que deverá sofrer mutação.
        pmut - probabilidade de ocorrer uma mutação.
    """
    def __init__(self, pmut, bits_por_intervalo):
        super(NTrocas, self).__init__(pmut)
        self.bits_por_intervalo = bits_por_intervalo

    def mutacao(self, ):
        """Alteração genética de membros da população usando dupla troca."""
        nmut = self.selecao()
        cromossos_totais = self.populacao[0].size

        if nmut.size != 0:
            intervalos = [[i, i + self.bits_por_intervalo]
                           for i in range(0,
                                          cromossos_totais,
                                          self.bits_por_intervalo
                                          )
                          ]
            ninter = len(intervalos)


            for i in nmut:
                inter1 = intervalos[randint(0, ninter)]
                inter2 = intervalos[randint(0, ninter)]

                while inter1 == inter2:
                    inter2 = intervalos[randint(0, ninter)]

                self.populacao[nmut,inter1[0]:inter1[1]], \
                self.populacao[nmut,inter2[0]:inter2[1]] = \
                self.populacao[nmut,inter2[0]:inter2[1]][::-1], \
                self.populacao[nmut,inter1[0]:inter1[1]][::-1]



#!/usr/bin/env python3.6
# -*- Coding: UTF-8 -*-
"""
Classe Abstrada de Seleção de Indivíduos para cruzamento.

Programa sob licença GNU V.3.
Desenvolvido por: E. S. Pereira.
Versão 0.0.1.
"""

from numpy.random import random
from numpy import array


class Selecao:
    """
    Seleciona indivíduos para cruzamento.
    Recebe como entrada:
        populacao - Objeto criado a partir da classe Populacao.
    """
    def __init__(self, populacao):
        self.populacao = populacao

    def selecionar(self, fitness=None):
        raise NotImplementedError("A ser implementado")

    def selecao(self, n, fitness=None):
        """
        Retorna uma população de tamanho n,
        selecionanda via roleta.
        """
        progenitores = array([self.selecionar(fitness)
                              for _ in range(n)])
        return self.populacao.populacao[progenitores]



#!/usr/bin/env python3.6
# -*- Coding: UTF-8 -*-
"""
Evolução.

Programa sob licença GNU V.3.
Desenvolvido por: E. S. Pereira.
Versão 0.0.1.
"""

from progressbar import progressbar
from numpy import array
from pathos.helpers import cpu_count, shutdown
from pathos.multiprocessing import ProcessPool

from .evolucao import Evolucao
from .populacao import Populacao


class AlgoritmoGenetico:
    def __init__(self, selecao, cruzamento, mutacao, bits, cromossomos,  **kwargs):
        self._bits = bits
        self._cromossomos = cromossomos

        self._parametros = {'tamanho_populacao': 100,
                            'nsele': 20,
                            'elitista': True,
                            'pcruz': 0.6,
                            'epidemia': 50,
                            'total_geracoes': 100,
                            'pmut': 0.05}

        for key in kwargs:
            if key not in self._parametros:
                raise NameError(f"Parametro invalido {key}")
            self._parametros[key] = kwargs[key]

        genes = self._bits * self._cromossomos 


        self._populacao = Populacao(self.avaliacao,
                                    genes, 
                                    self._parametros['tamanho_populacao'])
        self._cruzamento = cruzamento(self._parametros['tamanho_populacao'])

        self._mutacao = mutacao(pmut=self._parametros['pmut'])

        self._selecao = selecao(self._populacao)
        self._evolucao = Evolucao(self._populacao, self._selecao, self._cruzamento, self._mutacao)

        self._evolucao.nsele = self._parametros['nsele']
        self._evolucao.pcruz = self._parametros['pcruz']
        self._evolucao.manter_melhor = self._parametros['elitista']
        self._evolucao.epidemia = self._parametros['epidemia']

    def funcao_objetivo(self, individuo):
        raise NotImplementedError("Deve ser implementada")

    @property
    def parametros(self):
      return self._parametros

    def avaliacao(self, populacao):

        n = len(populacao)

        def steps(k):
            individuo = populacao[k, :]
            obj = self.funcao_objetivo(individuo)
            return obj

        ncpu = cpu_count()
        pool = ProcessPool(nodes=ncpu)
        pesos = array(pool.map(steps, range(n)))
        pool.close()
        pool.join()
        pool.clear()
        shutdown()

        return pesos

    def __call__(self):

        for _ in progressbar(range(self._parametros['total_geracoes'])):
            vmin, vmax = self._evolucao.evoluir()
            yield self._populacao.populacao[-1, :]


#!/usr/bin/env python3.6
# -*- Coding: UTF-8 -*-

from numpy import array


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def binarray2int(x):
    cnt = array([2 ** i for i in range(x.shape[1])])
    return array([(cnt * x[i,:]).sum()
                  for i in range(x.shape[0])]).reshape(1, x.shape[0])



#!/usr/bin/env python3.6
# -*- Coding: UTF-8 -*-
"""
Cruzamento por k-pontos.

Programa sob licença GNU V.3.
Desenvolvido por: E. S. Pereira.
Versão 0.0.1.
"""

from numpy.random import randint, random
from numpy import array

from .cruzamento import Cruzamento, NoCompatibleIndividualSize

class KPontos(Cruzamento):
    """
    Gerador de população via cruzamento usando o operador k-pontos.

    Entrada:
        tamanho_populacao - Tamanho final da população resultante.
    """
    def __init__(self, tamanho_populacao):
        super(KPontos, self).__init__(tamanho_populacao)

    def cruzamento(self, progenitor1, progenitor2):
        """
        Cruzamento via k-pontos de dois indivíduos.

        Entrada:
            ind1 - Primeiro indivíduo
            ind2 - Segundo indivíduo
        O tamanho de ambos os indivíduos deve ser igual, do contrário um erro
        será levantado.
        """
        n1 = len(progenitor1)
        n2 = len(progenitor2)

        if n1 != n2:
            msg = "Tamanho ind1 {0} diferente de ind2 {1}".format(n1, n2)
            raise NoCompatibleIndividualSize(msg)

        desc1 = progenitor1.copy()
        desc2 = progenitor2.copy()

        kp = randint(1, n1-2)

        k = []

        while len(k) <= kp:
            p = randint(1, n1-1)
            if p not in k:
                k.append(p)
        k.sort()

        troca = randint(0, 1)

        for i in range(kp):
            if troca == 0:
                troca = 1
            else:
                troca = 0
                desc1[k[i]:k[i + 1]] = progenitor2[k[i]:k[i + 1]]
                desc2[k[i]:k[i + 1]] = progenitor1[k[i]:k[i + 1]]

        return desc1, desc2



#!/usr/bin/env python3.6
# -*- Coding: UTF-8 -*-
"""
Roleta de Seleção de Indivíduos para cruzamento.

Programa sob licença GNU V.3.
Desenvolvido por: E. S. Pereira.
Versão 0.0.1.
"""

from numpy.random import choice
from numpy import array, where
from .selecao import Selecao


class Torneio(Selecao):
    """
    Seleciona indivíduos para cruzamento usando
    Torneio.
    Recebe como entrada:
        populacao - Objeto criado a partir da classe Populacao.
    """
    def __init__(self, populacao, tamanho=10):
        super(Torneio, self).__init__(populacao)
        self.tamanho = tamanho

    def selecionar(self, fitness):
        """Retorna o indivíduo campeão da rodada."""
        if fitness is None:
            fitness = self.populacao.avaliar()
        grupo = choice(fitness, size=self.tamanho)
        campeao = grupo.max()
        i = where(fitness == campeao)[0][0]
        return i
